from modulefinder import ModuleFinder
import os
import sys
import shutil
from shutil import copy
from pathlib import Path
import time
from textwrap import dedent
import platform

project_folder = Path.cwd()
project_name = project_folder.stem
build_folder = Path(project_folder / f'build_{platform.system()}')
build_folder.mkdir(exist_ok=True)

ignore_folders = []
ignore_filetypes = []

compressed_textures = []
compressed_textures_folder = Path(project_folder/'textures_compressed')
if compressed_textures_folder.exists():
    compressed_textures = compressed_textures_folder.iterdir()

def copytree(src, dst, symlinks=False, ignore_patterns=[], ignore_filetypes=[]):
    src = str(src)
    dst = str(dst)

    for item in os.listdir(src):
        if item in ignore_patterns:
            continue

        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            try:
                ignore_pattern = shutil.ignore_patterns(*(ignore_patterns + [f'*{e}' for e in ignore_filetypes]))
                shutil.copytree(s, d, symlinks, ignore_pattern)
            except Exception as e:
                print(e)
        else:
            if Path(s).suffix in ignore_filetypes:
                print('ignore filetype:', Path(s).suffix)
                continue
            if Path(s).stem in compressed_textures:
                continue
            shutil.copy2(s, d)



include_modules = []

python_dest = Path(build_folder / 'python')
python_dlls_dest = Path(build_folder / 'python/DLLs')
python_lib_dest = Path(build_folder / 'python/Lib')
src_dest = Path(build_folder / 'src')
build_engine = True
build_game = True
compile_to_pyc = True
entry_point = 'main.py'

for i, arg in enumerate(sys.argv):
    if arg == '--help':
        print(dedent('''
            package ursina application for windows10.
            provided with project folder path, creates a build folder where
            it copies python and project's dependent packages. requires a main.py file.
            copies game scripts and assets into 'build/src' folder.
            creates a .bat file to start the game.

            --ignore_folders=*      # add assets to ignore, for example: --ignore_folders=temp,unused
            --ignore_filetypes=*    # filetype to ignore, for example: --ignore_filetypes=.blend,.psd
            --name=''               # change project name
            --include_modules=*     # include extra modules like this: --include_modules=module_one,module_two,module_tree
            --overwrite             # don't ask to overwrite existing build, just overwrite
            --skip_engine
            --skip_game
            --compile_to_pyc=True/False

            Make sure to include any extra modules with --include_modules PIL,numpy for example.
            Any errors while the application is running will be logged in log.txt instead of the console.
            '''
            )
        )
        sys.exit()

    elif arg.startswith('--ignore_folders='):
        ignore_folders = arg.split('=')[1].split(',')
        print('ignoring folders:', ignore_folders)

    elif arg.startswith('--ignore_filetypes='):
        ignore_filetypes = arg.split('=')[1].split(',')
        print('ignoring filetypes:', ignore_filetypes)

    elif arg == '--name':
        project_name = sys.argv[i+1]

    elif arg.startswith('--include_modules='):
        include_modules = arg.split('=')[1].split(',')

    elif arg == '--skip_engine':
        build_engine = False
    elif arg == '--skip_game':
        build_game = False

    elif arg == '--compile_to_pyc=True':
        compile_to_pyc = True
    elif arg == '--compile_to_pyc=False':
        compile_to_pyc = False
    elif arg.startswith('--entry_point='):
        entry_point = arg.split('=')[1]


if (build_engine and python_dest.exists() or (build_game and src_dest.exists())):
    if '--overwrite' not in sys.argv:
        for e in (python_dest, src_dest):
            msg = f'Folder {e} already exists. \nProceed to delete and overwrite?'
            overwrite = input("%s (y/N) " % msg).lower() == 'y'
            # if not overwrite:
            #     print('stopped building')
            #     exit()
            if e == python_dest:
                build_engine = overwrite
            elif e == src_dest:
                build_game = overwrite


print('building project:', project_folder)
start_time = time.time()

if build_engine:
    if python_dest.exists():
        shutil.rmtree(str(python_dest))
    python_dest.mkdir()
    python_dlls_dest.mkdir()
    python_lib_dest.mkdir()


    # def copy_python():
    # copy files in python installation folder, but not the folders
    print('copying python')
    python_folder = Path(sys.executable).parent
    [copy(str(f), str(python_dest / f.name)) for f in python_folder.iterdir() if f.is_file()]


    # def copy_python_lib():
    print('copying python Lib files')
    [copy(str(f), str(python_lib_dest / f.name)) for f in Path(python_folder / 'Lib').iterdir() if f.is_file()]


    # def copy_always_included():
    print('copying always included modules and extra modules')
    include_modules = [f'Lib/site-packages/{e}' for e in include_modules]

    always_include = [
        'Lib/collections',
        'Lib/ctypes',
        'Lib/encodings',
        'Lib/importlib',
        'Lib/urllib',
        'Lib/logging',
        'Lib/xml',
        'Lib/re',
        'Lib/json',
        # 'Lib/site-packages/panda3d/',
        'Lib/site-packages/direct/',
        'Lib/site-packages/pyperclip/',
        'Lib/site-packages/PIL/',
        'Lib/site-packages/screeninfo/',
        'DLLs/libffi-7.dll',
        'DLLs/_ctypes.pyd',
        'DLLs',
        'Lib/site-packages/gltf',
        ]

    for path in always_include + include_modules:
        source = python_folder / path
        dest = python_dest / path
        print('copying always include:', source, '-->', dest)

        if source.is_file():
            dest.parent.mkdir(parents=True, exist_ok=True)
            copy(str(source), str(dest))
        elif source.is_dir():
            dest.mkdir(parents=True, exist_ok=True)
            copytree(source, dest)


    print('copying panda3d')
    (python_dest / 'Lib/site-packages/panda3d/').mkdir(parents=True, exist_ok=True)

    for f in (python_folder / 'Lib/site-packages/panda3d/').iterdir():
        if f.name in (
            'avcodec-55.dll',
            'fmodex64.dll',
            'egg.cp39-win_amd64.pyd',
            'libpandaode.dll',
            'ode.cp39-win_amd64.pyd',
            'models',
            'libp3tinydisplay.dll',
            'libp3assimp.dll',
            ):
            continue
        print('copying:', f, '-->', str(python_dest / 'Lib/site-packages/panda3d/' / f.name))
        if f.is_file():
            copy(str(f), str(python_dest / 'Lib/site-packages/panda3d/' / f.name))
        else:
            (python_dest / 'Lib/site-packages/panda3d/' / f.name).mkdir(parents=True, exist_ok=True)
            copytree(f, (python_dest / 'Lib/site-packages/panda3d/' / f.name))

    # def copy_ursina():
    print('copying ursina')
    import importlib
    spec = importlib.util.find_spec('ursina')
    ursina_path = Path(spec.origin).parent
    print('found ursina at:', ursina_path)
    dest = build_folder / 'python/Lib/site-packages/ursina'
    dest.mkdir(parents=True, exist_ok=True)
    copytree(ursina_path, dest, ignore_patterns=['samples', 'unused'], ignore_filetypes=['.blend', '.obj'])


    # print('copying found modules')
    # finder = ModuleFinder()
    # finder.run_script(str(project_folder) + '/main.py')

    # for name, mod in finder.modules.items():
    #     filename = mod.__file__
    #
    #     if filename is None:
    #         continue
    #     if '__' in name:
    #         print('ignore:', filename)
    #         continue
    #
    #     if 'Python' in filename and 'DLLs' in filename:
    #         print('copying:', filename)
    #         copy(filename, str(build_folder / 'python/DLLs'))
    #
    #     elif 'lib\\site-packages\\' in filename:
    #         print('copying module:', filename)
    #         forward_slashed = filename.split('lib\\site-packages\\')[1].replace('\\', '/')
    #         dir = build_folder / 'python/lib/site-packages' / forward_slashed
    #         dir.parent.mkdir(parents=True, exist_ok=True)
    #         copy(filename, dir)



if build_game:
    if src_dest.exists():
        shutil.rmtree(str(src_dest))
    src_dest.mkdir()

    ignore_folders.extend(['build_Windows', 'build_Linux', 'build.bat', '__pycache__', '.git'])
    ignore_filetypes.extend(['.gitignore', '.psd', '.zip'])

    if compile_to_pyc:
        import py_compile
        for f in project_folder.glob('**/*.py'):
            in_ignore = False
            for e in ignore_folders:
                if e in [str(e.name) for e  in f.parents]:
                    in_ignore = True
                    print('skip compiling:', f)
                    break
            if in_ignore:
                continue

            parents = f.relative_to(project_folder).parents
            if 'scenes' in parents:
                continue
            py_compile.compile(f, src_dest / (str(f)[len(str(project_folder))+1:]+'c'))

        ignore_filetypes.append('.py')


    print('copying assets')
    for f in project_folder.iterdir():
        name = f.name
        dest = Path(src_dest / f.name)
        if name in ignore_folders:
            print('ignore:', f)
            continue
        elif f.is_dir():
            print('copying assetfolder:', f, 'to', dest)
            dest.mkdir(parents=True, exist_ok=True)
            copytree(project_folder / f, dest, ignore_patterns=ignore_folders, ignore_filetypes=ignore_filetypes)
        elif f.is_file():
            if f.suffix in ignore_filetypes:
                continue
            print('copying asset:', f, 'to', src_dest / f.name)
            copy(str(f), str(dest))


    print('creating .bat file')
    c = ''
    if compile:
        c = 'c'
    with Path(build_folder / f'{project_name}.bat').open('w') as f:
        f.write(dedent(fr'''
            chcp 65001
            set PYTHONIOENCODING=utf-8

            call "python\python.exe" "src\{entry_point}{c}" > "log.txt" 2>&1'''
        ))

# make exe
# import subprocess
# import importlib
# spec = importlib.util.find_spec('ursina')
# ursina_path = Path(spec.origin).parent
# subprocess.call([
#     f'{ursina_path}\\scripts\\_bat_to_exe.bat',
#     f'{build_folder}\\{project_name}.bat',
#     f'\\build\\{project_folder.stem}.exe'
#     ])


print('build complete! time elapsed:', time.time() - start_time)



from ursina import *


class HealthBar(Button):

    def __init__(self, max_value=100, roundness=.25, animation_duration=.1, show_text=True, show_lines=False, **kwargs):
        super().__init__(position=(-.45*window.aspect_ratio,.45), origin=(-.5,.5), scale=(Text.size*20,Text.size), color=color.black66, highlight_color=color.black66, text='hp / max hp', radius=roundness, ignore=True)

        self.bar = Entity(parent=self, model=Quad(radius=roundness), origin=self.origin, z=-.01, color=color.red.tint(-.2), ignore=True)
        self.lines = Entity(parent=self.bar, y=-1, color=color.black33, ignore=True, enabled=show_lines, z=-.01)
        self.text_entity.scale *= .7

        self.max_value = max_value
        self.clamp = True
        self.roundness = roundness
        self.animation_duration = animation_duration
        self.show_lines = show_lines
        self.show_text = show_text

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.value = self.max_value
        self.text_entity.enabled = show_text


    @property
    def value(self):
        return self._value


    @value.setter
    def value(self, n):
        if self.clamp:
            n = clamp(n, 0, self.max_value)

        self._value = n

        self.bar.animate_scale_x(n/self.max_value, duration=self.animation_duration, curve=curve.in_out_bounce)
        self.text_entity.text = f'{n} / {self.max_value}'

        if self.lines.enabled:
            self.lines.model = Grid(n, 1)
            self.lines.origin = (-.5,-.5)

        if n / self.max_value >= self.scale_y / self.scale_x:
            aspect_ratio = n/self.max_value*self.scale_x / self.scale_y
            self.bar.model = Quad(radius=self.roundness, aspect=aspect_ratio)
        else:
            self.bar.model = 'quad'
        self.bar.origin = self.bar.origin



    @property
    def show_text(self):
        return self.text_entity.enabled
    @show_text.setter
    def show_text(self, value):
        self.text_entity.enabled = value
        print(self.text_entity.enabled)

    @property
    def show_lines(self):
        return self.lines.enabled
    @show_lines.setter
    def show_lines(self, value):
        self.lines.enabled = value

    @property
    def bar_color(self):
        return self.bar.color
    @bar_color.setter
    def bar_color(self, value):
        self.bar.color = value


    def __setattr__(self, name, value):
        if 'scale' and hasattr(self, 'model') and self.model:  # update rounded corners of background when scaling
            self.model.aspect = self.world_scale_x / self.world_scale_y
            self.model.generate()

        super().__setattr__(name, value)


if __name__ == '__main__':
    app = Ursina()

    health_bar_1 = HealthBar(bar_color=color.lime.tint(-.25), roundness=.5, value=50)
    print(health_bar_1.text_entity.enabled, health_bar_1.text_entity.text)
    # health_bar_1.show_text = False
    # health_bar_1.show_lines = True

    def input(key):
        if key == '+' or key == '+ hold':
            health_bar_1.value += 10
        if key == '-' or key == '- hold':
            health_bar_1.value -= 10
            print('ow')
    app.run()



'''Translated from https://github.com/AndrewRayCode/easing-utils/blob/master/src/easing.js'''

from math import cos, pi, sqrt, sin, asin, floor


def linear(t):
    return t


def in_sine(t):
    return -1 * cos(t * (pi / 2)) + 1


def out_sine(t):
    return sin(t * (pi / 2))


def in_out_sine(t):
    return -.5 * (cos(pi * t) - 1)


def in_quad(t):
    return t * t


def out_quad(t):
    return t * (2 - t)


def in_out_quad(t):
    if t < .5:
        return 2 * t * t
    else:
        return - 1 + (4 - 2 * t) * t


def in_cubic(t):
    return t * t * t


def out_cubic(t):
    t1 = t - 1
    return t1 * t1 * t1 + 1


def in_out_cubic(t):
    if t < .5:
        return 4 * t * t * t
    else:
        return (t - 1) * (2 * t - 2) * (2 * t - 2) + 1


def in_quart(t):
    return t * t * t * t


def out_quart(t):
    t1 = t - 1
    return 1 - t1 * t1 * t1 * t1


def in_out_quart(t):
    t1 = t - 1
    if t < .5:
        return 8 * t * t * t * t
    else:
        1 - 8 * t1 * t1 * t1 * t1


def in_quint(t):
    return t * t * t * t * t


def out_quint(t):
    t1 = t - 1
    return 1 + t1 * t1 * t1 * t1 * t1


def in_out_quint(t):
    t1 = t - 1
    if t < .5:
        return 16 * t * t * t * t * t
    else:
        return 1 + 16 * t1 * t1 * t1 * t1 * t1


def in_expo(t):
    return pow(2, 10 * (t - 1))


def out_expo(t):
    return -pow(2, -10 * t) + 1


def in_out_expo(t):
    scaledTime = t * 2
    scaledTime1 = scaledTime - 1

    if scaledTime < 1:
        return .5 * pow(2, 10 * scaledTime1)

    return .5 * (-pow(2, -10 * scaledTime1) + 2)


def in_circ(t):
    scaledTime = t / 1
    return -1 * (sqrt(1 - scaledTime * t) - 1)


def out_circ(t):
    t1 = t - 1
    return sqrt(1 - t1 * t1)


def in_out_circ(t):
    scaledTime = t * 2
    scaledTime1 = scaledTime - 2

    if scaledTime < 1:
        return -.5 * (sqrt(1 - scaledTime * scaledTime) - 1)

    return .5 * (sqrt(1 - scaledTime1 * scaledTime1) + 1)


def in_back(t, magnitude=1.70158):
    return t * t * ((magnitude + 1) * t - magnitude)


def out_back(t, magnitude=1.70158):
    scaledTime = (t / 1) - 1
    return (
                   scaledTime * scaledTime * ((magnitude + 1) * scaledTime + magnitude)
           ) + 1


def in_out_back(t, magnitude=1.70158):
    scaledTime = t * 2
    scaledTime2 = scaledTime - 2
    s = magnitude * 1.525

    if scaledTime < 1:
        return .5 * scaledTime * scaledTime * (
                ((s + 1) * scaledTime) - s
        )
    return .5 * (
            scaledTime2 * scaledTime2 * ((s + 1) * scaledTime2 + s) + 2
    )


def in_elastic(t, magnitude=.7):
    if t == 0 or t == 1:
        return t
    scaledTime = t / 1
    scaledTime1 = scaledTime - 1
    p = 1 - magnitude
    s = p / (2 * pi) * asin(1)

    return -(
            pow(2, 10 * scaledTime1) *
            sin((scaledTime1 - s) * (2 * pi) / p)
    )


def out_elastic(t, magnitude=.7):
    p = 1 - magnitude
    scaledTime = t * 2

    if t == 0 or t == 1:
        return t

    s = p / (2 * pi) * asin(1)
    return (
                   pow(2, -10 * scaledTime) *
                   sin((scaledTime - s) * (2 * pi) / p)
           ) + 1


def in_out_elastic(t, magnitude=0.65):
    p = 1 - magnitude
    if t == 0 or t == 1:
        return t

    scaledTime = t * 2
    scaledTime1 = scaledTime - 1
    s = p / (2 * pi) * asin(1)

    if scaledTime < 1:
        return -.5 * (
                pow(2, 10 * scaledTime1) *
                sin((scaledTime1 - s) * (2 * pi) / p)
        )

    return (
                   pow(2, -10 * scaledTime1) *
                   sin((scaledTime1 - s) * (2 * pi) / p) * .5
           ) + 1


def out_bounce(t):
    scaledTime = t / 1

    if scaledTime < 1 / 2.75:
        return 7.5625 * scaledTime * scaledTime

    elif scaledTime < 2 / 2.75:
        scaledTime2 = scaledTime - (1.5 / 2.75)
        return (7.5625 * scaledTime2 * scaledTime2) + .75

    elif scaledTime < 2.5 / 2.75:
        scaledTime2 = scaledTime - (2.25 / 2.75)
        return (7.5625 * scaledTime2 * scaledTime2) + 0.9375

    else:
        scaledTime2 = scaledTime - (2.625 / 2.75)
        return (7.5625 * scaledTime2 * scaledTime2) + 0.984375


def in_bounce(t):
    return 1 - out_bounce(1 - t)


def in_out_bounce(t):
    if t < .5:
        return in_bounce(t * 2) * .5

    return (out_bounce((t * 2) - 1) * .5) + .5


# generate boomeranged versions of all the functions
import sys
from textwrap import dedent

for e in dir(sys.modules[__name__]):
    item = getattr(sys.modules[__name__], e)
    if callable(item):
        exec(dedent(f'''
            def {e}_boomerang(t):
                if t < .5:
                    return {e}(t*2)
                else:
                    return {e}(1-((t-.5)*2))
        '''))


# bezier code is translated  from WebKit implementation
class CubicBezier:
    __slots__ = ['a', 'b', 'c', 'd', 'cx', 'bx', 'ax', 'cy', 'by', 'ay']

    def __init__(self, a, b, c, d):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        # pre-calculate the polynomial coefficients
        # irst and last control points are implied to be (0,0) and (1.0, 1.0)
        self.cx = 3.0 * a
        self.bx = 3.0 * (c - a) - self.cx
        self.ax = 1.0 - self.cx - self.bx

        self.cy = 3.0 * b
        self.by = 3.0 * (d - b) - self.cy
        self.ay = 1.0 - self.cy - self.by

    def sample_curve_x(self, t):
        return ((self.ax * t + self.bx) * t + self.cx) * t

    def sample_curve_y(self, t):
        return ((self.ay * t + self.by) * t + self.cy) * t

    def sample_curve_derivative_x(self, t):
        return (3.0 * self.ax * t + 2.0 * self.bx) * t + self.cx

    def calculate(self, x, epsilon=.0001):
        return self.sample_curve_y(self.solve_curve_x(x, epsilon))

    def solve_curve_x(self, t, epsilon=.0001):
        # First try a few iterations of Newton's method -- normally very fast.
        t0 = 0
        t1 = 0
        t2 = 0
        x2 = 0
        # d2 = 0
        # i = 0

        # t2 = t
        # for i in range(8):
        #     x2 = self.sample_curve_x(t2) - t
        #     if abs(x2) < epsilon:
        #         return t2
        #     d2 = self.sample_curve_derivative_x(t2)
        #     if abs(d2) < epsilon:
        #         break
        #
        #     t2 = t2 - x2 / d2

        # No solution found - use bi-section
        t0 = 0.0
        t1 = 1.0
        t2 = t

        if t2 < t0:
            return t0
        if t2 > t1:
            return t1

        while t0 < t1:
            x2 = self.sample_curve_x(t2)
            if abs(x2 - t) < epsilon:
                return t2
            if t > x2:
                t0 = t2
            else:
                t1 = t2

            t2 = (t1 - t0) * .5 + t0

        # Give up
        return t2


if __name__ == '__main__':
    '''Draws a sheet with every curve and its name'''
    from ursina import *
    from ursina import Ursina, camera, window, curve, Entity, Mesh, Text, color

    app = Ursina()
    camera.orthographic = True
    camera.fov = 16
    camera.position = (9, 6)
    window.color = color.black

    i = 0
    for e in dir(curve):
        try:
            item = getattr(curve, e)
            print(item.__name__, ':', item(.75))
            curve_renderer = Entity(
                model=Mesh(vertices=[Vec3(i / 31, item(i / 31), 0) for i in range(32)], mode='line', thickness=2),
                color=color.light_gray)
            row = floor(i / 8)
            curve_renderer.x = (i % 8) * 2.5
            curve_renderer.y = row * 1.75
            label = Text(parent=curve_renderer, text=item.__name__, scale=8, color=color.gray, y=-.1)
            i += 1
        except:
            pass

    c = CubicBezier(0, .5, 1, .5)
    print('-----------', c.calculate(.23))
    # for x, c in enumerate([curve.CubicBezier(0,1,1,0), curve.CubicBezier(0,.5,1,.5), curve.CubicBezier(0,0,0,0), curve.CubicBezier(0,.7,1,.3)]):
    #     verts = [Vec3(i/20, c.calculate(i/20), 0) for i in range(20)]
    #     Entity(parent=camera.ui, model=Mesh(vertices=verts, mode='line', thickness=3), scale=.25, x=-.5+(x*.25))

    window.exit_button.visible = False
    window.fps_counter.enabled = False
    app.run()
    '''
    These are used by Entity when animating, like this:

    e = Entity()
    e.animate_y(1, curve=curve.in_expo)

    e2 = Entity(x=1.5)
    e2.animate_y(1, curve=curve.CubicBezier(0,.7,1,.3))
    '''



from panda3d.core import TextNode
import builtins
# from direct.interval.IntervalGlobal import Sequence, Func, Wait, SoundInterval

import ursina
# from ursina import *
from ursina import camera
from ursina.entity import Entity
from ursina.sequence import Sequence, Func, Wait
from ursina import color
from ursina import destroy
# note:
# <scale:n> tag doesn't work well in the middle of text.
# only good for titles for now.

class Text(Entity):

    size = .025
    default_font = 'OpenSans-Regular.ttf'
    default_resolution = 1080 * size * 2
    start_tag = '<'
    end_tag = '>'

    def __init__(self, text='', start_tag=start_tag, end_tag=end_tag, ignore=True, **kwargs):
        super().__init__(ignore=ignore)
        self.size = Text.size
        self.parent = camera.ui

        self.setColorScaleOff()
        self.shader = None
        self.text_nodes = []
        self.images = []
        self.origin = (-.5, .5)

        self.font = Text.default_font
        self.resolution = Text.default_resolution
        self.line_height = 1
        self.use_tags = True
        self.start_tag = start_tag
        self.end_tag = end_tag
        self.text_colors = {'default' : color.text_color}

        for color_name in color.color_names:
            self.text_colors[color_name] = color.colors[color_name]

        self.tag = Text.start_tag+'default'+Text.end_tag
        self.current_color = self.text_colors['default']
        self.scale_override = 1
        self._background = None
        self.appear_sequence = None # gets created when calling appear()

        if 'origin' in kwargs:   # set the scale before model for correct corners
            setattr(self, 'origin', kwargs['origin'])
        if 'use_tags' in kwargs:
            setattr(self, 'use_tags', kwargs['use_tags'])

        if text != '':
            self.text = text

        for key, value in kwargs.items():
            if key == 'origin':
                continue
            setattr(self, key, value)


    @property
    def text(self):
        t = ''
        y = 0
        if self.text_nodes:
            y = self.text_nodes[0].getY()

        for tn in self.text_nodes:
            if y != tn.getY():
                t += '\n'
                y = tn.getY()

            t += tn.node().text

        return t


    @text.setter     # set this to update the text.
    def text(self, text):
        self.raw_text = text

        # clear stuff
        from ursina.ursinastuff import destroy  # needed to destroy inline images
        for img in self.images:
            destroy(img)
        self.images = []

        for tn in self.text_nodes:
            tn.remove_node()
        self.text_nodes = []

        # check if using tags
        if (not self.use_tags
            # or self.text == self.start_tag or self.text == self.end_tag
            # or not self.start_tag in text or not self.end_tag in text
            ):

            self.create_text_section(text)
            self.align()
            return

        # parse tags
        text = self.start_tag + self.end_tag + str(text) # start with empty tag for alignment to work?
        sections = []
        section = ''
        tag = self.start_tag+'default'+self.end_tag
        temp_text_node = TextNode('temp_text_node')
        temp_text_node.setFont(self.font)
        x = 0
        y = 0

        i = 0
        while i < len(text):
            char = text[i]
            if char == '\n':
                sections.append([section, tag, x, y])
                section = ''
                y -= 1
                x = 0
                i += 1

            elif char == self.start_tag: # find tag
                sections.append([section, tag, x, y])
                x += temp_text_node.calcWidth(section)
                section = ''

                tag = ''
                done = False
                for j in range(len(text)-i):
                    tag += text[i+j]
                    if text[i+j] == self.end_tag and len(tag) > 0:
                        i += j+1
                        done = True
                        break
                if not done:
                    i += 1
            else:
                section += char
                i += 1

        sections.append([section, tag, x, y])

        for i, s in enumerate(sections):
            tag = s[1]
            # move the text after image one space right
            if tag.startswith(self.start_tag+'image:'):
                for f in sections:
                    if f[3] == s[3] and f[2] > s[2]:
                        f[2] += .5

                s[2] += .5

            self.create_text_section(text=s[0], tag=s[1], x=s[2], y=s[3])

        self.align()


    def create_text_section(self, text, tag='', x=0, y=0):
        # print(text, tag)
        self.text_node = TextNode('t')
        self.text_node_path = self.attachNewNode(self.text_node)
        try:
            self.text_node.setFont(self._font)
        except:
            pass    # default font


        if tag != '<>':
            tag = tag[1:-1]

            if tag.startswith('hsb('):   # set color based on numbers
                tag = tag[4:-1]
                hsb_values = tuple(float(e.strip()) for e in tag.split(','))
                self.current_color = color.color(*hsb_values)

            elif tag.startswith('rgb('):   # set color based on numbers
                tag = tag[4:-1]
                rgb_values = (float(e.strip()) for e in tag.split(','))
                self.current_color = color.rgba(*rgb_values)

            if tag.startswith('scale:'):
                scale = tag.split(':')[1]
                self.scale_override = float(scale)

            elif tag.startswith('image:'):
                texture_name = tag.split(':')[1]
                image = Entity(
                    parent=self.text_node_path,
                    name='inline_image',
                    model='quad',
                    texture=texture_name,
                    color=self.current_color,
                    # scale=self.scale_override,
                    # position=(x*self.size*self.scale_override, y*self.size*self.line_height),
                    origin=(.0, -.25),
                    add_to_scene_entities=False,
                    )
                if not image.texture:
                    destroy(image)
                else:
                    self.images.append(image)

            else:
                if tag in self.text_colors:
                    self.current_color = self.text_colors[tag]

        self.text_node_path.setScale(self.scale_override * self.size)
        self.text_node.setText(text)
        self.text_node.setTextColor(self.current_color)
        self.text_node.setPreserveTrailingWhitespace(True)
        # self.text_node_path.setPos(
        #     x * self.size * self.scale_override,
        #     0,
        #     (y * self.size * self.line_height) - .75 * self.size)
        self.text_node_path.setPos(
            x * self.size * self.scale_override,
            (y * self.size * self.line_height) - .75 * self.size,
            0)
        self.text_nodes.append(self.text_node_path)

        return self.text_node

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, value):
        font = builtins.loader.loadFont(value)
        if font:
            self._font = font
            self._font.clear()  # remove assertion warning
            self._font.setPixelsPerUnit(self.resolution)
            self._font.setLineHeight(self.line_height)
            self.text = self.raw_text   # update tex

    @property
    def color(self): # sets the default color.
        return getattr(self, '_color', color.white)

    @color.setter
    def color(self, value):
        self._color = value
        self.current_color = value
        self.text_colors['default'] = value
        for tn in self.text_nodes:
            tn.node().setTextColor(value)
        for img in self.images:
            img.color = value


    @property
    def line_height(self):
        try:
            return self._line_height
        except:
            return 1

    @line_height.setter
    def line_height(self, value):
        self._line_height = value
        if self.use_tags:
            self.text = self.raw_text
        else:
            self._font.setLineHeight(value)

    @property
    def width(self): # gets the width of the widest line.
        if not hasattr(self, 'text'):
            return 0

        temp_text_node = TextNode('temp')
        temp_text_node.setFont(self._font)

        longest_line_length = 0
        for line in self.text.split('\n'):
            longest_line_length = max(longest_line_length, temp_text_node.calcWidth(line))

        return longest_line_length * self.size


    @property
    def height(self): # gets the height of the text
        return (len(self.lines) * self.line_height * self.size)

    @property
    def lines(self):
        return self.text.splitlines()

    @property
    def resolution(self):
        return self._font.getPixelsPerUnit()

    @resolution.setter
    def resolution(self, value):
        self._font.setPixelsPerUnit(value)

    @property
    def wordwrap(self): # set this to make the text wrap after a certain number of characters.
        if hasattr(self, '_wordwrap'):
            return self._wordwrap
        else:
            return 0

    @wordwrap.setter
    def wordwrap(self, value):
        self._wordwrap = value
        if not value:
            return

        new_text = ''
        x = 0
        for word in self.raw_text.replace(self.end_tag, self.end_tag+' ').split(' '):

            if word.startswith(self.start_tag) and new_text:
                new_text = new_text[:-1]

            if not word.startswith(self.start_tag):
                x += len(word) + 1

            if x >= value:
                new_text += '\n'
                x = 0

            new_text += word + ' '

        self.text = new_text


    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = value
        if self.text:
            self.text = self.raw_text

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, value):
        if value is True:
            self.create_background()
        elif self._background:
            from ursina.ursinastuff import destroy
            destroy(self._background)


    def align(self):
        value = self.origin

        linewidths = [self.text_nodes[0].node().calcWidth(line) for line in self.lines]
        # print('.........', linewidths)
        for tn in self.text_nodes:
            # center text horizontally
            # linenumber = abs(int(tn.getZ() / self.size / self.line_height))
            linenumber = abs(int(tn.getY() / self.size / self.line_height))
            tn.setX(tn.getX() - (linewidths[linenumber] / 2 * self.size * tn.getScale()[0] / self.size))
            # tn.setX(tn.getX() - (linewidths[linenumber] / 2 * self.size))
            # add offset based on origin/value
            # x -= half line width * text node scale
            tn.setX(
                tn.getX() - (linewidths[linenumber] / 2 * value[0] * 2 * self.size) * tn.getScale()[0] / self.size
                )
            # center text vertically
            halfheight = len(linewidths) * self.line_height / 2
            # tn.setZ(tn.getZ() + (halfheight * self.size))
            tn.setY(tn.getY() + (halfheight * self.size))
            # add offset
            # tn.setZ(tn.getZ() - (halfheight * value[1] * 2 * self.size))
            tn.setY(tn.getY() - (halfheight * value[1] * 2 * self.size))


    def create_background(self, padding=size*2, radius=size, color=ursina.color.black66):
        from ursina import Quad, destroy

        if self._background:
            destroy(self._background)

        self._background = Entity(parent=self, z=.01)

        if isinstance(padding, (int, float, complex)):
            padding = (padding, padding)

        w, h = self.width + padding[0], self.height + padding[1]
        self._background.x -= self.origin_x * self.width
        # if self.origin_x == .5:
        #     self._background.x += self.origin_x * self.width * 2
        self._background.y -= self.origin_y * self.height

        self._background.model = Quad(radius=radius, scale=(w, h))
        self._background.color = color


    def appear(self, speed=.025):
        self.enabled = True
        # self.visible = True   # setting visible seems to reset the colors
        if self.appear_sequence:
            self.appear_sequence.finish()

        self.appear_sequence = Sequence()
        for tn in self.text_nodes:
            target_text = tn.node().getText()
            tn.node().setText('')
            new_text = ''

            for char in target_text:
                new_text += char
                self.appear_sequence.append(Wait(speed))
                self.appear_sequence.append(Func(tn.node().setText, new_text))

        self.appear_sequence.start()
        return self.appear_sequence


    def get_width(string, font=None):
        t = Text(string)
        if font:
            t.font = font
        w = t.width
        from ursina import destroy
        destroy(t)
        return w



if __name__ == '__main__':
    from ursina import *
    from ursina import Ursina, dedent, window
    app = Ursina()
    # Text.size = .001
    descr = dedent('''
        Rainstorm
        Summon a rain storm to deal 5 water

        damage to everyone, test including yourself.
        1234 1234 1234 1234 1234 1234 2134 1234 1234 1234 1234 1234 2134 2134 1234 1234 1234 1234
        Lasts for 4 rounds.''').strip()

    # Text.default_font = 'VeraMono.ttf'
    # Text.default_font = 'consola.ttf'
    # color.text_color = color.lime
    Text.default_resolution = 1080 * Text.size
    test = Text(text=descr, wordwrap=30)
    # test.align()
    # test = Text(descr)

    # test.text = ''
    # print(test.images)
  # print('\n', test.text, '\n\n')
    # test.font = 'VeraMono.ttf'
    # Text.font = 'VeraMono.ttf'
    # test.origin = (.5, .5)
    # test.origin = (0, 0)
    # test.wordwrap = 40

    # text = Text(text=descr, wordwrap=10, origin=(-.5,.5), y=.25, background=True)
    # Entity(parent=camera.ui, model='circle', scale=.05, color=color.yellow, y=text.y, z=-1)


    def input(key):
        if key == 'a':
            # test.appear(speed=.025)
            print('a')
            test.text = '<default><image:file_icon> <red><image:file_icon> test '
            print('by', test.text)

    # test.create_background()
    window.fps_counter.enabled = False
    print('....', Text.get_width('yolo'))
    app.run()



import sys
import os
import math
import random
import inspect
import importlib
import subprocess
import time
from pathlib import Path

from ursina import application
from ursina.scene import instance as scene
from ursina.sequence import Sequence, Func, Wait


class Empty():
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key ,value)

class Default:
    pass


def invoke(function, *args, **kwargs):  # reserved keywords: 'delay', 'unscaled'
    delay = 0
    if 'delay' in kwargs:
        delay = kwargs['delay']
        del kwargs['delay']

    unscaled = False
    if 'unscaled' in kwargs:
        unscaled = kwargs['unscaled']
        del kwargs['unscaled']

    if not delay:
        function(*args, **kwargs)
        return None

    s = Sequence(
        Wait(delay),
        Func(function, *args, **kwargs)
    )
    s.unscaled = unscaled
    s.start()
    return s


def after(delay, unscaled=True):
    '''@after  decorator for calling a function after some time.

        example:
        @after(.4)
        def reset_cooldown():
            self.on_cooldown = False
            self.color = color.green
    '''
    def decorator(func):
        def wrapper(*args, **kwargs):
            invoke(func, *args, **kwargs, delay=delay, unscaled=unscaled)
        return wrapper()
    return decorator



def get_class_name(func):
    qualname_parts = func.__qualname__.split('.')
    class_name = qualname_parts[-2] if len(qualname_parts) > 1 else None
    return class_name

class every:
    '''@after  decorator for calling a function on an Entity after some time.

        example:
        @every(.1)
        def fixed_update():
            print('check collision')

        Using the @every decorator is the same as doing this in __init__() (on Entity):
        self.animations.append(Sequence(Func(self.fixed_update), Wait(.1), loop=True, started=True))
        The Sequence will call the function every .1 second, while adding it to
        self.animations ensures the Sequence gets cleaned up when the Entity gets destroyed.
    '''
    decorated_methods = []  # store decorated methods here

    def __init__(self, interval):
        self.interval = interval

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            # print(f"Calling {func.__name__} every {self.interval} seconds")
            return func(*args, **kwargs)

        wrapper._every = self  # add _every attribute to the decorated method
        wrapper._func = func
        every.decorated_methods.append(wrapper)  # store the decorated method
        return wrapper



def destroy(entity, delay=0):
    if delay == 0:
        _destroy(entity)
        return

    s = Sequence(Wait(delay), Func(_destroy, entity))
    s.start()
    return s

def _destroy(entity, force_destroy=False):
    from ursina import camera
    if not entity or entity == camera:
        return

    if entity.eternal and not force_destroy:
        return

    for child in entity.children:
        _destroy(child)

    if entity.collider:
        entity.collider.remove()

    if hasattr(entity, 'clip') and hasattr(entity, 'stop'): # stop audio
        entity.stop(False)

    if hasattr(entity, 'on_destroy'):
        entity.on_destroy()

    if entity in scene.entities:
        scene.entities.remove(entity)

    if entity in scene.collidables:
        scene.collidables.remove(entity)

    if hasattr(entity, '_parent') and entity._parent and hasattr(entity._parent, '_children') and entity in entity._parent._children:
        entity._parent._children.remove(entity)


    if hasattr(entity, 'scripts'):
        for s in entity.scripts:
            del s

    if hasattr(entity, 'animations'):
        for anim in entity.animations:
            anim.finish()
            anim.kill()

    if hasattr(entity, 'tooltip'):
        _destroy(entity.tooltip)

    if hasattr(entity, '_on_click') and isinstance(entity._on_click, Sequence):
        entity._on_click.kill()

    entity.removeNode()
    #unload texture
    # if hasattr(entity, 'texture') and entity.texture != None:
    #     entity.texture.releaseAll()

    del entity



def chunk_list(l, chunk_size):
    # yield successive chunks from list
    for i in range(0, len(l), chunk_size):
        yield l[i:i + chunk_size]


def flatten_list(l):
    # return [item for sublist in l for item in sublist]
    import itertools
    return list(itertools.chain(*l))

def flatten_completely(container):
    for i in container:
        if isinstance(i, (list,tuple)):
            for j in flatten(i):
                yield j
        else:
            yield i



def size_list():
    #return a list of current python objects sorted by size
    globals_list = []
    globals_list.clear()
    for e in globals():
        # object, size
        globals_list.append([e, sys.getsizeof(e)])
    globals_list.sort(key=operator.itemgetter(1), reverse=True)
    print('scene size:', globals_list)


def find_sequence(name, file_types, folders): # find frame_0, frame_1, frame_2 and so on
    for folder in folders:
        for file_type in file_types:
            files = list(folder.glob(f'**/{name}*.{file_type}'))
            if files:
                files.sort()
                return files
    return []


def import_all_classes(path=application.asset_folder, debug=False):
    path = str(path)
    sys.path.append(path)
    from ursina.string_utilities import snake_to_camel
    from glob import iglob
    imported_successfully = []

    for file_path in iglob(path + '**/*.py', recursive=True):
        if '\\build\\' in file_path or '__' in file_path:
            continue

        rel_path = file_path[len(path):][:-3].replace('\\', '.')
        if rel_path.startswith('.'):
            rel_path = rel_path[1:]
        module_name = os.path.basename(file_path).split('.')[0]
        class_name = snake_to_camel(module_name)
        module_name = module_name
        import_statement = 'from ' + rel_path + ' import *'

        try:
            exec(import_statement, globals())
            imported_successfully.append(module_name)
            if debug:
                print(import_statement)
        except:
            if debug:
                print('     x', import_statement)
            pass

    return imported_successfully


def print_on_screen(text, position=(0,0), origin=(-.5,.5), scale=1, duration=1):
    from ursina.text import Text
    text_entity = Text(text=text, position=position, origin=origin, scale=scale)
    destroy(text_entity, delay=duration)


class LoopingList(list):
    def __getitem__(self, i):
        return super().__getitem__(i % len(self))



if __name__ == '__main__':

    from ursina import *
    app = Ursina()



    # Player()
    app.run()



from main import *

app = Ursina()

from entity import *
from vec3 import Vec3
import color
from color import Color
from collider import Collider, BoxCollider
import time
from time import dt
from panda3d.bullet import BulletWorld, BulletBoxShape, BulletPlaneShape
from prefabs.editor_camera import EditorCamera

world = BulletWorld()
world.setGravity(Vec3(0, -9.81, 0))

cube = Entity(model = 'cube', physics_world = world, collider = 'box', y = 10, color = color.red)
#print(cube.physics_world)
#print(cube.collider)

ground = Entity(model = 'cube', physics_world = world, collider = 'box', position = (0, -5, 0), scale = (25, .1, 25), color = color.blue)

def update():
    dt = time.dt
    world.doPhysics(dt)

EditorCamera()

app.run()
